﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using System;
using System.IO;
using System.Xml.Linq;

#nullable disable
namespace HMI_Base
{
    public class PageSettings
    {
        // ページのタイトルを格納
        public string PageTitle { get; private set; } = "";

        // 背景色を格納
        public string BackgroundColor { get; private set; } = "";
        //背景画像
        public string BackgroundImage { get; set; } // wwwroot以下のパスを指定 /image/xxx.png
        //背景画像サイズ
        public string BackgroundSize { get; set; } // auto, cover, contain, 100px 200px（幅と高さをピクセル指定）, 50% 50%（幅と高さを親要素の50%指定）
        //背景画像ポジション
        public string BackgroundPosition { get; set; } // top left、center、bottom right, 10px 20px 50%など
        //ページの高さ ※自動でjsを使えば取得できるが処理が遅くなりそうなため固定とする
        public string PageHeight_Px { get; set; }

        public uint UpdateTime { get; private set; } = 1000;

        // xmlファイルの更新を監視するFileSystemWatcher
        private FileSystemWatcher _fileWatcher;

        // xmlファイルのパス（デフォルトは"HMI_PageSettings.xml"）
        private string _xmlFilePath = "HMI_PageSettings.xml";

        // UI更新を通知するイベント
        public event Func<Task> OnChange;

        // パラメータなしコンストラクタ（デフォルト設定で初期化）
        public PageSettings() { }

        // 初期化処理
        public void Initialize(string xmlFilePath)
        {
            // 引数で渡されたxmlファイルパスが空でない場合、そのパスを優先
            if (xmlFilePath.Length == 0)
                return;

            _xmlFilePath = xmlFilePath;

            // xml設定を読み込む
            LoadSettings();

            
            // xmlファイルの更新監視設定
            _fileWatcher = new FileSystemWatcher(Path.GetDirectoryName(_xmlFilePath))
            {
                Filter = Path.GetFileName(_xmlFilePath), // 指定されたファイルだけを監視
                NotifyFilter = NotifyFilters.LastWrite // 最終更新時の通知のみ
            };

            // ファイルが変更された際にOnFileChangedメソッドを呼び出す
            _fileWatcher.Changed += (s, e) => OnFileChanged();
            _fileWatcher.EnableRaisingEvents = true; // 更新監視を有効にする
        }

        // xmlファイルが変更された場合の処理
        private void OnFileChanged()
        {
            // 設定を再読み込み
            LoadSettings();

            // UIスレッドでUI更新を行うためにInvokeAsyncを使用
            if (OnChange != null)
            {
                // UI更新を非同期で行い、タスクの実行を待つ
                Task.Run(async () =>
                {
                    await OnChange.Invoke();
                });
            }
        }

        /// <summary>
        /// xmlファイルから設定を読み込み、各プロパティに設定を反映させる処理
        /// </summary>
        private void LoadSettings()
        {
            try
            {
                // xmlファイルが存在するか確認
                if (File.Exists(_xmlFilePath) == true)
                {
                    // xmlファイルを読み込む
                    var doc = XDocument.Load(_xmlFilePath);

                    // PageTitleをxmlから読み込む
                    PageTitle = doc.Root.Element("PageTitle")?.Value;

                    // BackgroundColorをxmlから読み込む
                    BackgroundColor = doc.Root.Element("BackgroundColor")?.Value;

                    var xmltime = doc.Root.Element("UpdateTime")?.Value;
                    if (xmltime != null && xmltime != "")//存在する場合のみ変換
                    {
                        //変換できた場合のみ格納 エラー時は1000
                        UpdateTime = UInt32.TryParse(xmltime, out uint temp) ? temp : 1000;
                    }
                    BackgroundColor = doc.Root.Element("BackgroundColor")?.Value;
                    PageHeight_Px = doc.Root.Element("PageHeight_Px")?.Value;
                    BackgroundImage = doc.Root.Element("BackgroundImage")?.Value;
                    BackgroundSize = doc.Root.Element("BackgroundSize")?.Value;
                    BackgroundPosition = doc.Root.Element("BackgroundPosition")?.Value;
                }
            }
            catch
            {
                // 例外が発生した場合は何もしない（エラーハンドリングなし）
            }
        }
    }
}
#nullable restore

