﻿////
/// Copyright(C) 2020 Micronet Co. All Rights Reserved.
//
using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Pipes;
using System.Linq;
using System.Text;
using System.Text.Json;
using System.Threading.Tasks;
using HMI_Container;

// .net core 3.0以上 または .NET 5以上 ここ以下の行はnull許容参照型が無効になる(従来通り)
#if NETCOREAPP || NET5_0_OR_GREATER
#nullable disable
#else
#endif
namespace HMI_Base
{
    /// <summary>
    /// HMIクライアントクラス
    /// </summary>
    public class HMI_Client
    {
        /// <summary>
        /// コンストラクタ
        /// </summary>
        public HMI_Client()
        {
            m_clMngMsgIn = new HMI_MngMsg();
            m_clMngMsgReadTag = new HMI_MngMsgReadTag();
            m_clMngMsgOut = new HMI_MngMsg();
            m_clMngMsgWriteTag = new HMI_MngMsgWriteTag();
        }

        /// <summary>
        /// デストラクタ
        /// </summary>
        ~HMI_Client()
        {
            Dispose();
        }

        /// <summary>
        /// 最新のエラーコード
        /// </summary>
        public UInt32 m_dwLastErr = (UInt32)HMI_Static.ErrCode.Ok;

        /// <summary>
        /// 最新のエラーコード
        /// </summary>
        public string m_srLastMsg = "";

        /// <summary>
        /// 名前付きパイプクラス：入力処理
        /// </summary>

        public NamedPipeClientStream m_clPipeIn = null;

        /// <summary>
        /// 名前付きパイプクラス：出力処理
        /// </summary>
        public NamedPipeClientStream m_clPipeOut = null;

        /// <summary>
        /// メッセージ通信クラス：入力処理
        /// </summary>
        public HMI_MngMsg m_clMngMsgIn = null;

        /// <summary>
        /// タグ読み込み管理クラス
        /// </summary>
        public HMI_MngMsgReadTag m_clMngMsgReadTag = null;
        
        /// <summary>
        /// メッセージ通信クラス：出力処理
        /// </summary>
        public HMI_MngMsg m_clMngMsgOut = null;

        /// <summary>
        /// タグ書き込み管理クラス
        /// </summary>
        public HMI_MngMsgWriteTag m_clMngMsgWriteTag = null;

        /// <summary>
        /// 名前付きパイプクラスのオープン
        /// </summary>
        /// <param name="srName">パイプ名</param>
        /// <param name="clPipe">パイプクラス</param>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        public void OpenPipe(string srName, ref NamedPipeClientStream clPipe, ref UInt32 dwRet, ref string srMsg)
        {
            dwRet = (UInt32)HMI_Static.ErrCode.Ok;
            srMsg = "";

            try
            {
                if (null != clPipe)
                {
                    // 既に生成済み
                    dwRet = (UInt32)HMI_Static.ErrCode.AlreadyCreate;
                    srMsg = "Already create pipe. " + srName;
                    return;
                }

                clPipe = new NamedPipeClientStream(".", srName, PipeDirection.InOut, PipeOptions.Asynchronous);

            }
            catch (Exception ex)
            {
                dwRet = (UInt32)ex.HResult;
                srMsg = ex.Message;
            }
        }

        /// <summary>
        /// 名前付きパイプクラスのクローズ
        /// </summary>
        /// <param name="clPipe">名前付きパイプクラス</param>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void ClosePipe(ref NamedPipeClientStream clPipe, ref UInt32 dwRet, ref string srMsg)
        {
            dwRet = (UInt32)HMI_Static.ErrCode.Ok;
            srMsg = "";

            if (null == clPipe)
            {
                // 既にクローズ済み
                dwRet = (UInt32)HMI_Static.ErrCode.AlreadyClose;
                srMsg = "Already close pipe. ";
            }
            else
            {
                clPipe.Close();
                clPipe = null;
            }
        }

        /// <summary>
        /// 名前付きパイプクラスの接続
        /// </summary>
        /// <param name="clPipe">パイプクラス</param>
        /// <param name="inTout">タイムアウト</param>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void ConnectPipe(ref NamedPipeClientStream clPipe, int inTout, ref UInt32 dwRet, ref string srMsg)
        {
            dwRet = (UInt32)HMI_Static.ErrCode.Ok;
            srMsg = "";

            try
            {
                if (null == clPipe)
                {
                    // 生成していない
                    dwRet = (UInt32)HMI_Static.ErrCode.NotCreate;
                    srMsg = "Not create pipe.";
                    return;
                }

                if (clPipe.IsConnected)
                {
                    // 既に接続済み
                    dwRet = (UInt32)HMI_Static.ErrCode.AlredyConnect;
                    srMsg = "Already connect pipe.";
                    return;
                }

                clPipe.Connect(inTout);
            }
            catch (System.Exception ex)
            {
                dwRet = (UInt32)ex.HResult;
                srMsg = ex.Message;
            }
        }

        /// <summary>
        /// 名前付きパイプの書き込み
        /// </summary>
        /// <param name="clPipe">パイプクラス</param>
        /// <param name="byData">書き込むデータ</param>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void WritePipe(ref NamedPipeClientStream clPipe, byte[] byData, ref UInt32 dwRet, ref string srMsg)
        {
            dwRet = (UInt32)HMI_Static.ErrCode.Ok;
            srMsg = "";

            try
            {
                if (null == clPipe)
                {
                    // 生成していない
                    dwRet = (UInt32)HMI_Static.ErrCode.NotCreate;
                    srMsg = "Not create pipe.";
                    return;
                }

                if (!clPipe.IsConnected)
                {
                    // 接続していない
                    dwRet = (UInt32)HMI_Static.ErrCode.NotConnect;
                    srMsg = "Not connect pipe.";
                    return;
                }

                clPipe.Write(byData);
            }
            catch (System.Exception ex)
            {
                dwRet = (UInt32)ex.HResult;
                srMsg = ex.Message;
            }
        }

        /// <summary>
        /// 名前付きパイプの読み込み 何バイトかわからないため用
        /// </summary>
        /// <param name="clPipe">パイプクラス</param>
        public const int timeoutMilliseconds = 1000;  // タイムアウト時間（ミリ秒）1000mmを設定 (仮)
        public static byte[] ReadPipeClientWithCancel(NamedPipeClientStream pipeClient)
        {
            using (MemoryStream ms = new MemoryStream())
            {
                byte[] buffer = new byte[1024];  // バッファは1024バイトに設定
                int bytesRead;
                while (pipeClient.IsConnected)
                {
                    // タスクを作成してReadを実行
                    Task<int> readTask = Task.Run(() =>
                    {
                        return pipeClient.Read(buffer, 0, buffer.Length);
                    });

                
                    // タイムアウトを設定
                    if (readTask.Wait(timeoutMilliseconds))
                    {
                        // タスクが完了した場合
                        bytesRead = readTask.Result;
                        if (bytesRead > 0)
                            ms.Write(buffer, 0, bytesRead);
                        else
                            break;
                    }
                    else
                    {
                        // タイムアウトが発生した場合
                        Console.WriteLine("Read operation timed out.");
                        // 必要に応じて中断処理を追加
                        break;
                    }
                    if (bytesRead < buffer.Length)//サイズが無い場合
                    {
                        // 受信バイトがバッファサイズ未満であれば終了
                        break;
                    }

                }

                // バイト配列に変換して返す
                return ms.ToArray();
            }
        }
        /// <summary>
        /// 名前付きパイプの読み込み
        /// </summary>
        /// <param name="clPipe">パイプクラス</param>
        /// <param name="byData">読み込んだデータ</param>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void ReadPipe(ref NamedPipeClientStream clPipe, ref byte[] byData, ref UInt32 dwRet, ref string srMsg)
        {
            dwRet = (UInt32)HMI_Static.ErrCode.Ok;
            srMsg = "";

            try
            {
                if (null == clPipe)
                {
                    // 生成していない
                    dwRet = (UInt32)HMI_Static.ErrCode.NotCreate;
                    srMsg = "Not create pipe.";
                    return;
                }

                if (!clPipe.IsConnected)
                {
                    // 接続していない
                    dwRet = (UInt32)HMI_Static.ErrCode.NotConnect;
                    srMsg = "Not connect pipe.";
                    return;
                }

                byData = ReadPipeClientWithCancel(clPipe);
                if( byData.Length == 0)
                {
                    dwRet = 0x1001;
                    srMsg = "データの取得に失敗しました。";
                }

            }
            catch (System.Exception ex)
            {
                dwRet = (UInt32)ex.HResult;
                srMsg = ex.Message;
            }
        }

        /// <summary>
        /// 破棄
        /// </summary>
        public void Dispose()
        {
            ClosePipe(ref m_clPipeIn, ref m_dwLastErr, ref m_srLastMsg);
            ClosePipe(ref m_clPipeOut, ref m_dwLastErr, ref m_srLastMsg);
        }

        /// <summary>
        /// 回線のオープン
        /// </summary>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 Open()
        {
            string srMsg = "";
            // 名前付きパイプクラス：入力処理のオープン
            m_dwLastErr = OpenIn(ref srMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            // 名前付きパイプクラス：出力処理のオープン
            m_dwLastErr = OpenOut(ref srMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }
        /// <summary>
        /// 回線のオープン
        /// </summary>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 OpenIn(ref string srMsg)
        {
            // 名前付きパイプクラス：入力処理のオープン
            OpenPipe(HMI_Static.HMI_PIPE_IN_NAME, ref m_clPipeIn, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }
            return m_dwLastErr;
        }
        public UInt32 OpenOut(ref string srMsg)
        {
            // 名前付きパイプクラス：出力処理のオープン
            OpenPipe(HMI_Static.HMI_PIPE_OUT_NAME, ref m_clPipeOut, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }
            return m_dwLastErr;
        }

        /// <summary>
        /// 回線のクローズ
        /// </summary>
        /// <returns></returns>
        public UInt32 Close()
        {
            // 名前付きパイプクラス：入力処理のクローズ
            m_dwLastErr = CloseIn();
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            // 名前付きパイプクラス：出力処理のクローズ
            m_dwLastErr = CloseOut();
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }
        /// <summary>
        /// 回線のクローズ
        /// </summary>
        /// <returns></returns>
        public UInt32 CloseIn()
        {
            // 名前付きパイプクラス：入力処理のクローズ
            ClosePipe(ref m_clPipeIn, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }
            return m_dwLastErr;
        }
        /// <summary>
        /// 回線のクローズ
        /// </summary>
        /// <returns></returns>
        public UInt32 CloseOut()
        {
            // 名前付きパイプクラス：出力処理のクローズ
            ClosePipe(ref m_clPipeOut, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }
        /// <summary>
        /// サーバーへ接続
        /// </summary>
        /// <param name="inTout">接続タイムアウト(ms)</param>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 Connect(int inTout)
        {
            // 名前付きパイプクラス：入力処理の接続
            ConnectPipe(ref m_clPipeIn, inTout, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            // 名前付きパイプクラス：出力処理の接続
            ConnectPipe(ref m_clPipeOut, inTout, ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }

        /// <summary>
        /// サーバーへ接続
        /// </summary>
        /// <param name="inTout">接続タイムアウト(ms)</param>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 ConnectIn(int inTout, ref string srMsg)
        {
            // 名前付きパイプクラス：入力処理の接続
            ConnectPipe(ref m_clPipeIn, inTout, ref m_dwLastErr, ref srMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }
            return m_dwLastErr;
        }
        /// <summary>
        /// サーバーへ接続
        /// </summary>
        /// <param name="inTout">接続タイムアウト(ms)</param>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 ConnectOut(int inTout, ref string srMsg)
        {
            ConnectPipe(ref m_clPipeOut, inTout, ref m_dwLastErr, ref srMsg);
            if (m_dwLastErr != (UInt32)HMI_Static.ErrCode.Ok)
            {
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }


        /// <summary>
        /// タグ読み込み要求伝文(データ部)：タグ情報追加
        /// </summary>
        /// <param name="srName">タグ名</param>
        public void AddReqReadTagDataTag(string srName)
        {
            //クリアしてから追加
            m_clMngMsgReadTag.ClearReqDataTags();
            m_clMngMsgReadTag.AddReqDataTag(srName);
        }
        public void AddReqReadTagDataTags(List<string> Names)
        {
            //クリアしてから追加
            m_clMngMsgReadTag.ClearReqDataTags();
            foreach(string srName in Names)
                m_clMngMsgReadTag.AddReqDataTag(srName);
        }

        public void ClearReqReadTagData()
        {
            m_clMngMsgReadTag.ClearReqDataTags();
        }
        /// <summary>
        /// タグ読み込み処理
        /// </summary>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 EntryReadTag()
        {
            // タグ読み込み要求処理
            EntryReqReadTag(ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != 0)
            {
                // タグ読み込み要求失敗のため終了
                return m_dwLastErr;
            }

            // タグ読み込み応答処理
            EntryResReadTag(ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != 0)
            {
                // タグ読み込み要求失敗のため終了
                return m_dwLastErr;
            }

            return m_dwLastErr;
        }

        /// <summary>
        /// タグ読み込み要求処理
        /// </summary>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void EntryReqReadTag(ref UInt32 dwRet, ref string srMsg)
        {
            UInt32 dwRetL = 0;
            string srMsgL = "";
            int inRet = 0;

            dwRet = 0;
            srMsg = "";

            // 要求伝文(データ部)をシリアライズする
            m_clMngMsgReadTag.SerializeReqData();

            // 要求伝文(ヘッダ部)の作成
            m_clMngMsgIn.m_clReqHead.UpdateHead((UInt16)HMI_Msg.MsgNo.ReqReadTag, (UInt32)m_clMngMsgReadTag.m_byReqData.Length);

            // 要求伝文(ヘッダ部)をシリアライズする
            m_clMngMsgIn.m_clReqHead.SerializeHead();

            // 要求伝文(ヘッダ部)の送信
            WritePipe(ref m_clPipeIn, m_clMngMsgIn.m_clReqHead.m_byHead, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)の送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ReqReadTag Head]" + srMsgL;
                return;
            }

            // 要求伝文(ヘッダ部)ACKの受信
            ReadPipe(ref m_clPipeIn, ref m_clMngMsgIn.m_clReqHeadAck.m_byAck, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ReqReadTag HeadAck]" + srMsgL;
                return;
            }

            // 要求伝文(ヘッダ部)ACKをデシリアライズする
            m_clMngMsgIn.m_clReqHeadAck.DeserializeAck();

            // 要求伝文(ヘッダ部)ACK判定
            inRet = m_clMngMsgIn.m_clReqHeadAck.JudgeAck();
            if (inRet != 0)
            {
                dwRet = (UInt32)inRet;
                srMsg = "[Judge ReqReadTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }

            // エラーコード判定
            if (m_clMngMsgIn.m_clReqHeadAck.m_clAck.m_dwStatus != 0) {
                dwRet = m_clMngMsgIn.m_clReqHeadAck.m_clAck.m_dwStatus;
                srMsg = "[Judge ReqReadTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return; 
            }

            // 要求伝文(データ部)の送信
            WritePipe(ref m_clPipeIn, m_clMngMsgReadTag.m_byReqData, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(データ部)の送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ReqReadTag Data]" + srMsgL;
                return;
            }

            // 要求伝文(データ部)ACKの受信
            ReadPipe(ref m_clPipeIn, ref m_clMngMsgIn.m_clReqDataAck.m_byAck, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(データ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ReqReadTag DataAck]" + srMsgL;
                return;
            }

            // 要求伝文(データ部)ACKをデシリアライズする
            m_clMngMsgIn.m_clReqDataAck.DeserializeAck();

            // 要求伝文(ヘッダ部)ACK判定
            inRet = m_clMngMsgIn.m_clReqDataAck.JudgeAck();
            if (inRet != 0)
            {
                dwRet = (UInt32)inRet;
                srMsg = "[Judge ReqReadTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }

            // エラーコード判定
            if (m_clMngMsgIn.m_clReqDataAck.m_clAck.m_dwStatus != 0)
            {
                dwRet = m_clMngMsgIn.m_clReqDataAck.m_clAck.m_dwStatus;
                srMsg = "[Judge ReqReadTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }


            //
            // ここから応答
            //

            // 応答伝文(ヘッダ部)をシリアライズする
            m_clMngMsgIn.m_clResHead.SerializeHead();

            //応答伝文(ヘッダ部)の受信
            ReadPipe(ref m_clPipeIn, ref m_clMngMsgIn.m_clResHead.m_byHead, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                //応答伝文(ヘッダ部)の受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ResReadTag Head]" + srMsgL;
                return;
            }

            // 要求伝文(ヘッダ部)をデシリアライズする
            m_clMngMsgIn.m_clResHead.DeserializeHead();

            // 応答伝文(ヘッダ部)判定
            inRet = m_clMngMsgIn.m_clResHead.JudgeHead();
            
            // 応答伝文(ヘッダ部)ACKの作成
            m_clMngMsgIn.m_clResHeadAck.UpdateAck(m_clMngMsgIn.m_clResHead.m_clHead.m_woMsgNo, (UInt32)inRet);

            // 応答伝文(ヘッダ部)ACKをシリアライズ
            m_clMngMsgIn.m_clResHeadAck.SerializeAck();

            //応答伝文(ヘッダ部)ACKの送信
            WritePipe(ref m_clPipeIn, m_clMngMsgIn.m_clResHeadAck.m_byAck, ref dwRetL, ref srMsgL);
            if (inRet != 0)
            {
                // 応答伝文(ヘッダ部)判定結果NG
                dwRet = dwRetL;
                srMsg = "[Judge ReqReadTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }
            if (dwRetL != 0)
            {
                //応答伝文(ヘッダ部)ACKの送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ResReadTag HeadAck]" + srMsgL;
                return;
            }

            //応答伝文(データ部)の受信
            ReadPipe(ref m_clPipeIn, ref m_clMngMsgReadTag.m_byResData, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                //応答伝文(データ部)の受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ResReadTag Data]" + srMsgL;
                return;
            }

            // 応答伝文(データ部)をデシリアライズする
            m_clMngMsgReadTag.DeserializeResData();

            // 応答伝文(データ部)ACKの作成
            m_clMngMsgIn.m_clResDataAck.UpdateAck(m_clMngMsgIn.m_clResHead.m_clHead.m_woMsgNo, (UInt32)inRet);

            //応答伝文(データ部)ACKをシリアライズする
            m_clMngMsgIn.m_clResDataAck.SerializeAck();

            //応答伝文(データ部)ACKの送信
            WritePipe(ref m_clPipeIn, m_clMngMsgIn.m_clResDataAck.m_byAck, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                //応答伝文(データ部)ACKの送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ResReadTag DataAck]" + srMsgL;
                return;
            }
        }

        /// <summary>
        /// タグ読み込み応答処理
        /// </summary>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void EntryResReadTag(ref UInt32 dwRet, ref string srMsg)
        {
            UInt32 dwRetL = 0;
            string srMsgL = "";

            dwRet = 0;
            srMsg = "";

            // 応答伝文(ヘッダ部)の受信
            //
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ResReadTag Head]" + srMsgL;
                return;
            }

            // 応答伝文(ヘッダ部)ACKの送信
            //
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ResReadTag HeadAck]" + srMsgL;
                return;
            }

            // 応答伝文(データ部)の受信
            //
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ResReadTag Data]" + srMsgL;
                return;
            }

            // 応答伝文(データ部)ACKの送信
            //
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ResReadTag DataAck]" + srMsgL;
                return;
            }
        }

        /// <summary>
        /// タグ書き込み要求伝文(データ部)：タグ情報追加
        /// </summary>
        /// <param name="srName">タグ名</param>
        public void AddReqWriteTagDataTag(string srName, UInt16 woType, object val)
        {
            m_clMngMsgWriteTag.ClearReqDataTag();
            m_clMngMsgWriteTag.AddReqDataTag(srName, woType, val);
        }
        public void AddReqWriteTagDataTags(List<MsgDataTagValue> tags)
        {
            m_clMngMsgWriteTag.ClearReqDataTag();
            foreach(MsgDataTagValue tag in tags)
                m_clMngMsgWriteTag.AddReqDataTag(tag.m_srName, tag.m_woType, tag.m_obVal);
        }

        /// <summary>
        /// タグ書き込み処理
        /// </summary>
        /// <returns>0=正常終了/それ以外=エラー</returns>
        public UInt32 EntryWriteTag()
        {
            // タグ書き込み要求処理
            EntryReqWriteTag(ref m_dwLastErr, ref m_srLastMsg);
            if (m_dwLastErr != 0)
            {
                // タグ読み込み要求失敗のため終了
                return m_dwLastErr;
            }

            // タグ書き込み応答処理は無し
            return m_dwLastErr;
        }

        /// <summary>
        /// タグ書き込み要求処理
        /// </summary>
        /// <param name="dwRet">戻り値</param>
        /// <param name="srMsg">メッセージ</param>
        private void EntryReqWriteTag(ref UInt32 dwRet, ref string srMsg)
        {
            UInt32 dwRetL = 0;
            string srMsgL = "";
            int inRet = 0;

            dwRet = 0;
            srMsg = "";

            // 要求伝文(データ部)をシリアライズする
            m_clMngMsgWriteTag.SerializeReqData();

            // ★TODO データサイズは正しいサイズが書き込めない(シリアライズしたデータのサイズになるので)どうするべきか要検討 いらないかも？
            // 要求伝文(ヘッダ部)の作成
            m_clMngMsgOut.m_clReqHead.UpdateHead((UInt16)HMI_Msg.MsgNo.ReqWriteTag, (UInt32)m_clMngMsgWriteTag.m_byReqData.Length);

            // 要求伝文(ヘッダ部)をシリアライズする
            m_clMngMsgOut.m_clReqHead.SerializeHead();

            // 要求伝文(ヘッダ部)の送信
            WritePipe(ref m_clPipeOut, m_clMngMsgOut.m_clReqHead.m_byHead, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)の送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ReqWriteTag Head]" + srMsgL;
                return;
            }

            // 要求伝文(ヘッダ部)ACKの受信
            ReadPipe(ref m_clPipeOut, ref m_clMngMsgOut.m_clReqHeadAck.m_byAck, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ReqWriteTag HeadAck]" + srMsgL;
                return;
            }

            // 要求伝文(ヘッダ部)ACKをデシリアライズする
            m_clMngMsgOut.m_clReqHeadAck.DeserializeAck();

            // 要求伝文(ヘッダ部)ACK判定
            inRet = m_clMngMsgOut.m_clReqHeadAck.JudgeAck();
            if (inRet != 0)
            {
                dwRet = (UInt32)inRet;
                srMsg = "[Judge ReqWriteTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }

            // エラーコード判定
            if (m_clMngMsgOut.m_clReqHeadAck.m_clAck.m_dwStatus != 0)
            {
                dwRet = m_clMngMsgOut.m_clReqHeadAck.m_clAck.m_dwStatus;
                srMsg = "[Judge ReqWriteTag HeadAck]" + " ErrCode[" + dwRet.ToString() + "]";
                return;
            }

            // 要求伝文(データ部)の送信
            WritePipe(ref m_clPipeOut, m_clMngMsgWriteTag.m_byReqData, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)の送信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Send ReqWriteTag Data]" + srMsgL;
                return;
            }

            // 要求伝文(データ部)ACKの受信
            ReadPipe(ref m_clPipeOut, ref m_clMngMsgOut.m_clReqDataAck.m_byAck, ref dwRetL, ref srMsgL);
            if (dwRetL != 0)
            {
                // 要求伝文(ヘッダ部)ACKの受信失敗のため終了
                dwRet = dwRetL;
                srMsg = "[Recv ReqWriteTag DataAck]" + srMsgL;
                return;
            }

        }
    }
}
// .net core 3.0以上 または .NET 5以上 ここ以下の行はnull許容参照型が有効になる
#if NETCOREAPP || NET5_0_OR_GREATER
#nullable restore
#else
#endif
